package es.uca.xmleye.gui.trees;

/*
 Copyright (C) 2005-2008 Antonio Garcia Dominguez

 This program is free software; you can redistribute it 
 and/or modify it under the terms of the GNU General Public 
 License as published by the Free Software Foundation; 
 either version 2 of the License, or (at your option) any 
 later version.

 This program is distributed in the hope that it will be 
 useful, but WITHOUT ANY WARRANTY; without even the implied 
 warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 PURPOSE. See the GNU General Public License for more 
 details.

 You should have received a copy of the GNU General Public 
 License along with this program; if not, write to the Free 
 Software Foundation, Inc., 59 Temple Place, Suite 330, 
 Boston, MA 02111-1307 USA
 */

import org.w3c.dom.*;
import javax.swing.*;
import javax.swing.tree.*;

import es.uca.xmleye.SearchException;
import es.uca.xmleye.util.UnaryOperator;

/**
 * Especialización de JTree para visualización de árboles DOM XML.
 */
public class DOMTree extends JTree {
	/**
	 * 
	 */
	private static final long serialVersionUID = -2285899598562281752L;

	/**
	 * Crea un nuevo árbol que muestra el modelo <code>d</code>.
	 */
	public DOMTree(DOMTreeModel d) {
		super(d);
		getSelectionModel().setSelectionMode(
				TreeSelectionModel.SINGLE_TREE_SELECTION);
		setShowsRootHandles(true);
	}

	/* METODOS DE NAVEGACIÓN */

	/**
	 * Selecciona un nodo identificado por una ruta XPath <code>sXPath</code>.
	 * 
	 * @param sXPath
	 *            Cadena con la ruta XPath.
	 * @throws XPathExpressionException
	 *             La ruta XPath no es válida.
	 */
	public void selectNode(final String sXPath) throws SearchException {
		selectPath(new UnaryOperator<TreePath, TreePath>() {
			public TreePath ejecutar(TreePath t) {
				try {
					return ((DOMTreeModel) getModel()).getPathToNode(sXPath);
				} catch (SearchException exB) {
					exB.printStackTrace();
					return null;
				}
			}
		});
	}

	/**
	 * Selecciona un nodo contenido en el árbol.
	 * 
	 * @param nodo
	 *            Nodo perteneciente al modelo del árbol.
	 */
	public void selectNode(final Node nodo) {
		selectPath(new UnaryOperator<TreePath, TreePath>() {
			public TreePath ejecutar(TreePath t) {
				return ((DOMTreeModel) getModel()).nodePath(nodo);
			}
		});
	}

	/**
	 * Cambia el nodo seleccionado al padre del nodo actual, si lo tiene. De lo
	 * contrario, no hace nada.
	 */
	public void selectParent() {
		selectPath(new UnaryOperator<TreePath, TreePath>() {
			public TreePath ejecutar(TreePath t) {
				return ((DOMTreeModel) getModel()).getPathToParent(t);
			}
		});
	}

	/**
	 * Cambia el nodo seleccionado al primer hijo del nodo actual, si lo tiene.
	 * De lo contrario, no hace nada.
	 */
	public void selectFirstChild() {
		selectPath(new UnaryOperator<TreePath, TreePath>() {
			public TreePath ejecutar(TreePath t) {
				return ((DOMTreeModel) getModel()).getPathToFirstChild(t);
			}
		});
	}

	/**
	 * Cambia el nodo seleccionado al siguiente hermano del nodo actual, si lo
	 * tiene. De lo contrario, no hace nada.
	 */
	public void selectNextSibling() {
		selectPath(new UnaryOperator<TreePath, TreePath>() {
			public TreePath ejecutar(TreePath t) {
				return ((DOMTreeModel) getModel()).getPathToNextSibling(t);
			}
		});
	}

	/**
	 * Cambia el nodo seleccionado al anterior hermano del nodo actual, si lo
	 * tiene. De lo contrario, no hace nada.
	 */
	public void selectPreviousSibling() {
		selectPath(new UnaryOperator<TreePath, TreePath>() {
			public TreePath ejecutar(TreePath t) {
				return ((DOMTreeModel) getModel()).getPathToPreviousSibling(t);
			}
		});
	}

	/* MÉTODOS DE IMPLEMENTACIÓN */

	/**
	 * Cambia la selección a la que le indique la operación
	 * <code>fNuevoCamino</code>.
	 */
	private void selectPath(UnaryOperator<TreePath, TreePath> fNuevoCamino) {
		final TreeSelectionModel selArb = getSelectionModel();

		final TreePath antiguoCamino = selArb.getSelectionPath();
		final TreePath nuevoCamino = fNuevoCamino.ejecutar(antiguoCamino);
		if (nuevoCamino != null) {
			selArb.setSelectionPath(nuevoCamino);
			scrollPathToVisible(nuevoCamino);
		}
	}

}
