package es.uca.xmleye.presentation;

/*
 Copyright (C) 2008 Antonio García Domínguez

 This program is free software; you can redistribute it 
 and/or modify it under the terms of the GNU General Public 
 License as published by the Free Software Foundation; 
 either version 2 of the License, or (at your option) any 
 later version.

 This program is distributed in the hope that it will be 
 useful, but WITHOUT ANY WARRANTY; without even the implied 
 warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 PURPOSE. See the GNU General Public License for more 
 details.

 You should have received a copy of the GNU General Public 
 License along with this program; if not, write to the Free 
 Software Foundation, Inc., 59 Temple Place, Suite 330, 
 Boston, MA 02111-1307 USA
 */

import java.io.File;
import java.util.ArrayList;
import java.util.Locale;
import java.util.ResourceBundle;
import javax.swing.tree.*;
import org.w3c.dom.*;

import es.uca.xmleye.Editor;
import es.uca.xmleye.ParsingException;
import es.uca.xmleye.SearchException;
import es.uca.xmleye.EditException;
import es.uca.xmleye.DocumentModel;
import es.uca.xmleye.SearchRequest;
import es.uca.xmleye.gui.trees.DataDrivenTreeModel;
import es.uca.xmleye.gui.util.XHTMLSearchKeyHighlighter;
import es.uca.xmleye.util.ValueWrapper;
import es.uca.xmleye.util.FileChangePublisher;
import es.uca.xmleye.xslt.TransformationException;
import es.uca.xmleye.xslt.PreprocessingStylesheet;
import es.uca.xmleye.xslt.ViewStylesheet;

/**
 * Modelo de presentación de un documento.
 */
public class DocumentPresentationModel {
	// Expresión regular para marcado de apariciones de la clave de búsqueda
	private final static String DESTACA_REGEXP = "<font class=\"searchkey\">$1</font>";
	// Recursos para localización
	private static final ResourceBundle miscRes;

	static {
		miscRes = ResourceBundle.getBundle("es.uca.xmleye.l10n.MiscMessages",
				Locale.getDefault());
	}

	private WndMainPresentationModel ventanaPrincipal; // Modelo de
	// ventana
	// principal
	private ViewStylesheet hojaVista; // Stylesheet de visualización
	private DocumentModel modDocumento; // Modelo del documento

	private boolean bNavegable; // ¿Documento navegable?
	private DataDrivenTreeModel modArbol; // Modelo del árbol para JTree
	private TreePath seleccion; // Ruta seleccionada
	private String fuenteXHTML; // Código XHTML a visualizar

	private SearchRequest ultimaBusqueda; // Última búsqueda realizada
	private NodeList resUltimaBusqueda; // Caché de resultados
	private int indiceUltimoResultado; // Indice del último resultado

	private FileChangePublisher notCambios; // Notificador de cambios de

	// la fuente

	/* --- CONSTRUCTORES --- */

	/**
	 * Constructor para el caso en que no hay documento abierto.
	 */
	public DocumentPresentationModel(WndMainPresentationModel ventanaPrincipal)
			throws ParsingException, TransformationException {
		this.ventanaPrincipal = ventanaPrincipal;
		this.hojaVista = null;
		this.modDocumento = null;

		this.bNavegable = false;
		this.modArbol = new DataDrivenTreeModel();
		this.seleccion = null;

		this.ultimaBusqueda = null;
		this.resUltimaBusqueda = null;

		this.fuenteXHTML = generarFuenteXHTML(this.seleccion);
		this.notCambios = null;
	}

	public DocumentPresentationModel(DocumentModel pMod,
			ViewStylesheet pHojaVista, WndMainPresentationModel ventanaPrincipal)
			throws TransformationException {

		this.ventanaPrincipal = ventanaPrincipal;
		this.hojaVista = pHojaVista;

		setModeloDocumento(pMod);

		this.bNavegable = true;
		this.ultimaBusqueda = null;
		this.resUltimaBusqueda = null;

		// Si se quiere recibir notificaciones, habrá que suscribirse
		this.notCambios = new FileChangePublisher(pMod.getFicheroOriginal(),
				FileChangePublisher.INTERVALO_OMISION,
				FileChangePublisher.INACTIVO);
	}

	/* --- METODOS PUBLICOS --- */

	// Accesores para sincronización
	public String getNombreHojaVista() {
		return (hojaVista != null) ? hojaVista.getNombre()
				: DefaultPreferences.INICIAL_VISTA_OMISION;
	}

	public String getNombreHojaPreproc() {
		return (modDocumento != null) ? modDocumento.getHojaPreproc()
				.getNombre() : DefaultPreferences.INICIAL_PREPROCESADO_OMISION;
	}

	public String getNombreDocumento() {
		return (modDocumento != null) ? getFicheroOrigen().getName() : null;
	}

	public TreeModel getModArbol() {
		return modArbol;
	}

	public TreePath getSeleccion() {
		return seleccion;
	}

	public String getFuenteXHTML() {
		return fuenteXHTML;
	}

	public boolean getNavegable() {
		return bNavegable;
	}

	public FileChangePublisher getNotCambios() {
		return notCambios;
	}

	public String getRutaBase() {
		return (modDocumento != null) ? modDocumento.getRutaBase() : null;
	}

	public synchronized NodeList solicitudBusqueda(SearchRequest e)
			throws SearchException {
		final ValueWrapper<String> elem = ventanaPrincipal.inicioTarea(miscRes
				.getString("status_searching"));

		try {
			final Node nodoSel;
			if (getSeleccion().getLastPathComponent() != null) {
				nodoSel = (Node) getSeleccion().getLastPathComponent();
			} else {
				nodoSel = (Node) modArbol.getRoot();
			}

			NodeList res;
			int posSel = buscarPosicion(nodoSel, resUltimaBusqueda);
			if (e.equals(ultimaBusqueda) && posSel != -1) {
				this.indiceUltimoResultado = posSel + 1;
				res = this.resUltimaBusqueda;
			} else {
				this.ultimaBusqueda = e;
				this.resUltimaBusqueda = null;
				this.indiceUltimoResultado = 0;

				res = modDocumento.buscar(e, nodoSel);
			}

			return res;
		} catch (SearchException ex) {
			throw ex;
		} finally {
			ventanaPrincipal.finTarea(elem);
		}
	}

	public synchronized NodeList repetirBusqueda() {
		return this.resUltimaBusqueda;
	}

	public synchronized void finBusqueda(NodeList resultados, SearchRequest e) {
		// Sólo aceptamos los resultados de la búsqueda más reciente
		// (diferenciamos entre referencias distintas - posiblemente
		// tengan orígenes distintos)
		if (e != null && !e.equals(ultimaBusqueda))
			return;

		// Almacenamos los resultados por si se repite la búsqueda
		resUltimaBusqueda = resultados;

		// Cambiamos la selección al primer/siguiente resultado (si lo hay)
		if (resultados != null && resultados.getLength() > 0) {
			if (indiceUltimoResultado >= resultados.getLength()) {
				// Volvemos al primer resultado si no quedan más
				indiceUltimoResultado = 0;
			}
			this.seleccion = modArbol.nodePath(resultados
					.item(indiceUltimoResultado++));
		}
	}

	public synchronized String solicitudCambioSeleccion(TreePath nuevaSel)
			throws TransformationException {
		final ValueWrapper<String> elem = ventanaPrincipal.inicioTarea(miscRes
				.getString("status_transforming"));

		try {
			String res = generarFuenteXHTML(nuevaSel);
			return res;
		} catch (TransformationException ex) {
			throw ex;
		} finally {
			ventanaPrincipal.finTarea(elem);
		}

	}

	public synchronized void finCambioSeleccion(String nuevoXHTML,
			TreePath nuevaSel) {
		this.seleccion = nuevaSel;
		this.fuenteXHTML = nuevoXHTML;
	}

	public synchronized void editarDocumento() throws EditException {
		File f = getFicheroOrigen();
		ArrayList<String> orden = ventanaPrincipal.getOrdenEditor(f);

		if (f != null) {
			Editor.lanzarEditor(f, orden);
		}
	}

	/* --- METODOS DE PAQUETE: SÓLO USAR EN WndMainPresentationModel --- */

	// / Obtención del fichero XML
	synchronized File getFicheroFuente() {
		return this.modDocumento.getFicheroXML();
	}

	// / Obtención del fichero origen
	synchronized File getFicheroOrigen() {
		return this.modDocumento.getFicheroOriginal();
	}

	/**
	 * Notificación por WndMainPresentationModel de cambio de hoja de
	 * visualización.
	 */
	synchronized void cambioHojaVisualizacion(ViewStylesheet nuevaHoja)
			throws TransformationException {
		this.hojaVista = nuevaHoja;
		this.fuenteXHTML = generarFuenteXHTML(this.seleccion);
	}

	/**
	 * Notificación por WndMainPresentationModel de cambio de hoja de
	 * preprocesado.
	 */
	synchronized void cambioHojaPreprocesado(PreprocessingStylesheet nuevaHoja)
			throws ParsingException, TransformationException {
		modDocumento.setHojaPreprocesado(nuevaHoja);
		setModeloDocumento(this.modDocumento);
	}

	// / Notificación por DocumentPresentationModel de cambio de importacion
	// automatica
	synchronized void cambioAutoImportacion(boolean activo) {
		if (activo) {
			notCambios.reanudar();
		} else {
			notCambios.pausar();
		}
	}

	// / Obtención de modelo de documento
	synchronized DocumentModel getModeloDocumento() {
		return this.modDocumento;
	}

	// / Cambio de modelo de documento
	synchronized void setModeloDocumento(DocumentModel pMod)
			throws TransformationException {

		this.modDocumento = pMod;
		this.modArbol = new DataDrivenTreeModel(modDocumento.getArbolDOM());

		try {
			TreePath rootElementPath = modArbol.getPathToNode("/*[1]");
			if (rootElementPath != null) {
				this.seleccion = rootElementPath;
			} else {
				this.seleccion = new TreePath(modArbol.getRoot());
			}
		} catch (SearchException ex) {
		}

		this.fuenteXHTML = generarFuenteXHTML(this.seleccion);
	}

	/* --- METODOS PRIVADOS --- */
	private String generarFuenteXHTML(TreePath sel)
			throws TransformationException {
		if (sel != null && (Element) sel.getLastPathComponent() != null) {
			final Element nodoSel = (Element) sel.getLastPathComponent();
			String xhtml = hojaVista.generarXHTML(nodoSel, modDocumento
					.getArbolDOM());

			if (resUltimaBusqueda != null
					&& buscarPosicion(nodoSel, resUltimaBusqueda) != -1) {
				xhtml = new XHTMLSearchKeyHighlighter().marcarApariciones(
						xhtml, ultimaBusqueda.getClaveBusqueda(),
						DESTACA_REGEXP, ultimaBusqueda.getPalabraCompleta(),
						ultimaBusqueda.getCoincidenMayusMinus());
			}

			return xhtml;
		} else {
			return "<html><head></head><body></body></html>";
		}
	}

	/**
	 * Devuelve la posición de aguja en pajar o -1 si no está.
	 */
	private int buscarPosicion(Node aguja, NodeList pajar) {
		if (pajar == null || aguja == null)
			return -1;

		for (int i = 0; i < pajar.getLength(); ++i) {
			if (aguja.equals(pajar.item(i))) {
				return i;
			}
		}
		return -1;
	}

	/**
	 * Cierra el modelo de presentación del documento, interrumpiendo cualquier
	 * proceso que pueda quedarse ejecutándose indefinidamente de lo contrario.
	 */
	public void cerrar() {
		notCambios.interrumpir();
	}
}
