package es.uca.xmleye.xslt;

/*
 Copyright (C) 2008 Antonio García Domínguez

 This program is free software; you can redistribute it 
 and/or modify it under the terms of the GNU General Public 
 License as published by the Free Software Foundation; 
 either version 2 of the License, or (at your option) any 
 later version.

 This program is distributed in the hope that it will be 
 useful, but WITHOUT ANY WARRANTY; without even the implied 
 warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 PURPOSE. See the GNU General Public License for more 
 details.

 You should have received a copy of the GNU General Public 
 License along with this program; if not, write to the Free 
 Software Foundation, Inc., 59 Temple Place, Suite 330, 
 Boston, MA 02111-1307 USA
 */

import java.io.File;
import java.io.FilenameFilter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.ResourceBundle;
import java.util.Locale;

import javax.xml.transform.URIResolver;

import es.uca.xmleye.util.BinaryOperator;

/**
 * <p>
 * Esta clase permite el soporte de extensibilidad e internacionalización de los
 * mecanismos de transformación XML-&gt;XML y XML-&gt;HTML usados por el
 * programa mediante hojas de estilo definidas por el usuario.
 * </p>
 * 
 * <p>
 * Por lo tanto, es responsable de conocer dónde se hallan en el sistema de
 * ficheros las hojas.
 * </p>
 * 
 * <p>
 * Los resolvedores usados en las hojas reconocen de entre las URI especiales
 * siguientes las de su tipo (VISTA o PREPROCESADO)
 * </p>
 * 
 * <ul>
 * <li><em>current_view (URI_VISTA_ACTUAL)</em>:
 * <p>
 * obtiene la hoja de personalización de vista seleccionada actualmente por el
 * usuario.
 * </p>
 * </li>
 * 
 * <li><em>current_preproc (URI_PREPROCESADO_ACTUAL)</em>:
 * <p>
 * obtiene la hoja de personalización de preprocesado seleccionada actualmente
 * por el usuario.
 * </p>
 * </li>
 * 
 * <li><em>view_X (VIEW_URI_PREFIX + "X")</em>:
 * <p>
 * obtiene la hoja de personalización de vista con el nombre lógico "X". Así
 * podemos extender hojas de visualización ya existentes.
 * </p>
 * 
 * <p>
 * Intenta usar aquella que se ajuste mejor a la local por defecto (sigue un
 * mecanismo de resolución de ficheros idéntico al que sigue
 * ResourceBundle.getBundle).
 * </p>
 * </li>
 * 
 * <li><em>preproc_X (PREPROC_URI_PREFIX + "X")</em>:
 * <p>
 * obtiene la hoja de personalización de preprocesado con el nombre lógico X.
 * Establece los mismos mecanismos de internacionalización y extensibilidad que
 * el anterior.
 * </p>
 * </li>
 * </ul>
 */
public class StylesheetRepository {
	// Usado para internacionalizar los mensajes de error
	protected static final ResourceBundle errRes;

	static {
		errRes = ResourceBundle.getBundle("es.uca.xmleye.l10n.ErrorMessages",
				Locale.getDefault());
	}

	// Ruta relativa de carpeta con hojas XSLT a partir del directorio de xmleye
	private String dirHojas;

	// Hojas de preprocesado disponibles (según nombre)
	private Map<String, PreprocessingStylesheet> hojasPreprocesado;
	// Hojas de visualización disponibles (según nombre)
	private Map<String, ViewStylesheet> hojasVista;

	/* --- CONSTRUCTORES --- */

	/**
	 * Inicializa un repositorio bajo un cierto directorio.
	 * 
	 * @param pDirHojas
	 *            Ruta absoluta o relativa al directorio desde el que se lanzó
	 *            XMLEye del directorio base.
	 * @throws StylesheetRepositoryException
	 *             Fallo durante la inicialización del repositorio.
	 */
	public StylesheetRepository(String pDirHojas)
			throws StylesheetRepositoryException {
		this.dirHojas = pDirHojas;

		// Lista de nombres de hojas de vista
		hojasVista = crearMapaHojas(StylesheetConstants.VISTA,
				new BinaryOperator<String, URIResolver, ViewStylesheet>() {
					public ViewStylesheet ejecutar(String hoja,
							URIResolver resolvedor) {
						return new ViewStylesheet(getHojaPrincipalVista(),
								resolvedor, hoja);
					}
				});

		hojasPreprocesado = crearMapaHojas(
				StylesheetConstants.PREPROCESADO,
				new BinaryOperator<String, URIResolver, PreprocessingStylesheet>() {
					public PreprocessingStylesheet ejecutar(String hoja,
							URIResolver resolvedor) {
						return new PreprocessingStylesheet(
								getHojaPrincipalPreprocesado(), resolvedor,
								hoja);
					}
				});
	}

	/* --- METODOS PUBLICOS --- */

	/**
	 * Obtiene la hoja de visualización con nombre lógico <code>nombre</code>,
	 * o devuelve <code>null</code> si no está disponible.
	 */
	public ViewStylesheet getHojaVisualizacion(String nombre) {
		return hojasVista.get(nombre);
	}

	/**
	 * Obtiene la hoja de preprocesado con nombre lógico <code>nombre</code>,
	 * o devuelve <code>null</code> si no está disponible.
	 */
	public PreprocessingStylesheet getHojaPreprocesado(String nombre) {
		return hojasPreprocesado.get(nombre);
	}

	/**
	 * Obtiene la lista de nombres de hoja de visualización.
	 */
	public List<String> getNombresHojasVisualizacion() {
		return new ArrayList<String>(hojasVista.keySet());
	}

	/**
	 * Obtiene la lista de nombres de hoja de preprocesado.
	 */
	public List<String> getNombresHojasPreprocesado() {
		return new ArrayList<String>(hojasPreprocesado.keySet());
	}

	/* --- METODOS PRIVADOS --- */

	/**
	 * Crea un mapa con todas las hojas de un determinado tipo. Usa un functor
	 * para crear la hoja.
	 * 
	 * @throws StylesheetRepositoryException
	 *             Alguna de las hojas esperadas no se ha encontrado.
	 */
	private <T extends Stylesheet> Map<String, T> crearMapaHojas(String tipo,
			BinaryOperator<String, URIResolver, T> fCrearHoja)
			throws StylesheetRepositoryException {
		String[] nombresHojasVista = buscarHojasDisponibles(tipo);
		Map<String, T> mapaHojas = new HashMap<String, T>();

		for (String hoja : nombresHojasVista) {
			URIResolver resolvedor = crearResolvedor(hoja, tipo,
					nombresHojasVista);
			mapaHojas.put(hoja, fCrearHoja.ejecutar(hoja, resolvedor));
		}

		return mapaHojas;
	}

	/**
	 * Crea un resolvedor para una hoja con nombre <code>nombreHoja</code> del
	 * tipo <code>tipoHoja</code>, posiblemente accediendo a otras hojas del
	 * mismo tipo con nombres <code>restoHojas</code>.
	 * 
	 * @throws StylesheetRepositoryException
	 *             Fallo al localizar alguno de los ficheros.
	 * @see StylesheetConstants.VISTA
	 * @see StylesheetConstants.PREPROCESADO
	 */
	private URIResolver crearResolvedor(String nombreHoja, String tipoHoja,
			String[] restoHojas) throws StylesheetRepositoryException {
		/* Directorio base de esta hoja */
		String dirBase = dirHojas + File.separator + tipoHoja + File.separator
				+ nombreHoja;

		HashURIResolver res = new HashURIResolver(new BaseDirectoryURIResolver(
				dirBase, new BaseDirectoryURIResolver(dirHojas,
						new NoOpURIResolver())));

		String actualURI; // URI que representa la hoja en cuestión
		String prefijoURI; // URI que aparece como prefijo de las demás
		if (tipoHoja.equals(StylesheetConstants.VISTA)) {
			actualURI = StylesheetConstants.URI_VISTA_ACTUAL;
			prefijoURI = StylesheetConstants.VIEW_URI_PREFIX;
		} else {
			actualURI = StylesheetConstants.URI_PREPROCESADO_ACTUAL;
			prefijoURI = StylesheetConstants.PREPROC_URI_PREFIX;
		}

		final File fActual = localizarFichero(nombreHoja, tipoHoja);
		res.establecer(actualURI, fActual);
		for (String hoja : restoHojas) {
			final File fConPrefijo = localizarFichero(hoja, tipoHoja);
			res.establecer(prefijoURI + hoja, fConPrefijo);
		}

		return res;
	}

	/**
	 * Busca las hojas de usuario disponibles para selección del usuario desde
	 * el directorio base especificado de un determinado tipo.
	 * 
	 * @throws StylesheetRepositoryException
	 *             El directorio base con las hojas no existe.
	 */
	private String[] buscarHojasDisponibles(String tipo)
			throws StylesheetRepositoryException {
		final File dirBase = new File(dirHojas + File.separator + tipo);

		if (!dirBase.exists()) {
			throw new StylesheetRepositoryException(String.format(errRes
					.getString("extensions_notfound"), tipo));
		}

		return dirBase.list(new FilenameFilter() {
			public boolean accept(File f, String nombre) {
				return f.isDirectory();
			}
		});
	}

	/**
	 * <p>
	 * Resuelve la referencia a la hoja de personalización con nombre lógico
	 * <code>nombreLogHoja</code> y tipo <code>tipo</code>.
	 * </p>
	 * 
	 * <p>
	 * Realiza la resolución de local, buscando igual que lo hace
	 * ResourceBundle: si se da que la local por defecto es es_ES (idioma_país),
	 * la hoja se llama "foo" y es de tipo "vista", probará con:
	 * </p>
	 * 
	 * <ol>
	 * <li>(dir. de hojas "vista")/foo/foo_es_ES.xsl (personalización a nivel
	 * de país+lenguaje)</li>
	 * <li>(dir. de hojas "vista")/foo/foo_es.xsl (personalización a nivel de
	 * lenguaje)</li>
	 * <li>(dir. de hojas "vista")/foo/foo.xsl (por defecto)</li>
	 * </ol>
	 * 
	 * <p>
	 * Todo esto se hace relativo al directorio de trabajo.
	 * </p>
	 * 
	 * @param nombreLogHoja
	 *            El nombre lógico (sin atender a local) de la hoja.
	 * @param tipo
	 *            El tipo de la hoja: <code>VISTA</code> o
	 *            <code>PREPROCESADO</code>.
	 * @return Archivo con la hoja XSLT exacta que hay que invocar, o
	 *         <code>null</code> si no se encuentra.
	 * @throws StylesheetRepositoryException
	 *             El directorio de la hoja no existe.
	 */
	private File localizarFichero(String nombreLogHoja, String tipo)
			throws StylesheetRepositoryException {
		final String baseURL = dirHojas + File.separator + tipo
				+ File.separator + nombreLogHoja;

		/* Comprobamos que existe la hoja con dicho nombre lógico */
		final File dirHoja = new File(baseURL);
		if (!dirHoja.exists()) {
			throw new StylesheetRepositoryException(String.format(errRes
					.getString("customss_notfound"), nombreLogHoja));
		}

		/* Buscamos siguiendo el orden de resolución de local */
		final Locale localOmision = Locale.getDefault();
		final String leng = localOmision.getLanguage();
		final String pais = localOmision.getCountry();

		/* Probamos con país y lenguaje */
		File hoja = new File(dirHoja, nombreLogHoja + "_" + leng + "_" + pais
				+ ".xsl");
		if (hoja.exists())
			return hoja;

		/* Probamos con lenguaje */
		hoja = new File(dirHoja, nombreLogHoja + "_" + leng + ".xsl");
		if (hoja.exists())
			return hoja;

		/* Probamos la hoja por defecto */
		hoja = new File(dirHoja, nombreLogHoja + ".xsl");
		if (hoja.exists())
			return hoja;

		throw new StylesheetRepositoryException(String.format(errRes
				.getString("customss_notfound"), nombreLogHoja));
	}

	String getHojaPrincipalVista() {
		return dirHojas + File.separator
				+ StylesheetConstants.VIEW_MAIN_FILENAME;
	}

	String getHojaPrincipalPreprocesado() {
		return dirHojas + File.separator
				+ StylesheetConstants.PREPROC_MAIN_FILENAME;
	}
}
